package BadTraffic2::FlatFile;

use warnings;
use strict;
use Carp;
use File::Copy qw(move);

# exports
use base qw( Exporter );
our @EXPORT      = qw( );
our @EXPORT_OK   = qw( );
our %EXPORT_TAGS = (
    ALL => [ @EXPORT, @EXPORT_OK ],    # if :ALL requested
);

=head1 NAME

BadTraffic2::FlatFile - Firewall handler for flat files

=head1 SYNOPSIS

This code is used internally.
  
=head1 DESCRIPTION

Example of a BadTraffic2 firewall handler.

=head1 INTERFACE 

There are no public methods.

=cut

=head2 new

Initialize the handler.

=cut

sub new {
    my ( $class, %parameters ) = @_;

    # defaults and valid keys
    my $self = bless(
        {   FW_NAME => '/tmp/badtraffic_list.txt',
            DEBUG   => 0,
        },
        ref($class) || $class
    );

    foreach my $key ( sort keys %parameters ) {
        if ( defined $self->{$key} ) {
            $self->{$key} = $parameters{$key};
        }
        else {
            # quietly ignore extra params
        }
    }

    # tests to make sure the handler can be used
    croak "FW_TYPE flatfile requires FW_NAME be set to a file path"
        if ( !$self->{FW_NAME} );
    print STDERR "BadTraffic2: Using firewall flatfile $self->{FW_NAME}\n"
        if ( $self->{DEBUG} );

    # tell upstream code that we prefer to rewrite the entire list
    $self->{_UPDATES} = 'all';

    return ($self);
}

=head2 config_check 

Check and fix firewall configuration

=cut

sub config_check {
    my $self = shift;

}

=head2 read_set

Read list of blocked addresses as a set object.

=cut 

sub read_set {
    my $self = shift;
    my $raw  = shift;    # if true, return an array of entries

    my $cidrset = Net::CIDR::Set->new( { type => 'ipv4' } );
    my $ct      = 0;
    my $rawlist = [];
    if ( open( my $in, '<', $self->{FW_NAME} ) ) {
        while ( my $lin = <$in> ) {
            chomp $lin;
            $lin =~ s{#.*\z}{};           # clean comments
            $lin =~ s{\A\s+|\s+\z}{}g;    # clean blankspace
            next if ( $lin =~ m{\A\s*\z} );    # skip lines with no content
            if ( $lin =~ $BadTraffic2::CIDRregex ) {
                $cidrset->add($lin);
                push @$rawlist, $lin;
                $ct++;
            }
            else {
                carp "Skipped bad CIDR: $lin" if ( $self->{DEBUG} );
            }
        }
        close($in);
    }
    elsif ( -s $self->{FW_NAME} ) {
        croak "flatfile $self->{FW_NAME} is not readable";
    }
    print STDERR "flatfile: read $ct entries\n" if ( $self->{DEBUG} );
    return ( $raw ? $rawlist : $cidrset );
}

=head2 write_set 

Write entire list of addresses, removing previous contents.

=cut

sub write_set {
    my $self    = shift;
    my $cidrset = shift;

    my $ct  = 0;
    my $cur = $self->{FW_NAME};
    my $new = $cur . '.new';
    open( my $out, '>', $new ) || croak "Cannot write new flatfile";
    my $iter = $cidrset->iterate_cidr;
    while ( my $entry = $iter->() ) {
        $entry .= '/32' if ( $entry !~ m{/} );    # fix single IPs to be /32s
        print $out "$entry\n";
        $ct++;
    }
    close($out);
    print STDERR "flatfile: write $ct entries\n" if ( $self->{DEBUG} );
    move( $new, $cur ) || croak "Failed to replace flatfile";
}

=head2 add_entry 

Add a single ip/cidr rule to the existing list.

=cut

sub add_entry {
    my $self   = shift;
    my $ipcidr = shift;

    # lazy add: append the entry to the end of the current file
    # assumes the input has already be filtered for a proper string
    open( my $out, '>>', $self->{FW_NAME} )
        || croak "Cannot append to flatfile";
    print $out "$ipcidr\n";
    close($out);
    print STDERR "flatfile: added $ipcidr\n" if ( $self->{DEBUG} );
}

=head2 del_entry

Remove a single ip/cidr rule from the existing list.

=cut 

sub del_entry {
    my $self   = shift;
    my $ipcidr = shift;

    # ipcidr must be an exact match for a file entry
    $ipcidr .= '/32' if ( $ipcidr !~ m{/} );
    my $ct  = 0;
    my $cur = $self->{FW_NAME};
    my $new = $cur . '.new';
    if ( open( my $in, '<', $cur ) ) {
        open( my $out, '>', $new ) || croak "Cannot write new flatfile";
        while ( my $lin = <$in> ) {
            chomp $lin;
            if ( $lin =~ m{\A\s*$ipcidr\s*\z} ) {
                $ct++;
            }
            else {
                print $out "$lin\n";
            }
        }
        close($out);
        close($in);
        move( $new, $cur ) || croak "Failed to replace flatfile";
    }
    if ($ct) {
        print STDERR "flatfile: removed $ipcidr\n" if ( $self->{DEBUG} );
    }
    else {
        print STDERR "flatfile: remove $ipcidr not found\n"
            if ( $self->{DEBUG} );
    }
}

=head2 commit

Update synchronize all changes

=cut

sub commit {
    my $self = shift;
}

1;    # Magic true value required at end of module
__END__

=head1 AUTHOR

Charles J. Brabec  C<< <brabec@ncsu.edu> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2019, Charles J. Brabec C<< <brabec@ncsu.edu> >>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
