package BadTraffic2::IPSet;

use warnings;
use strict;
use Carp;

# exports
use base qw( Exporter );
our @EXPORT      = qw( );
our @EXPORT_OK   = qw( );
our %EXPORT_TAGS = (
    ALL => [ @EXPORT, @EXPORT_OK ],    # if :ALL requested
);

# this should match the settings used by the ipset module in puppet
my $SetType = 'hash:net family inet hashsize 8192 maxelem 131072';

=head1 NAME

BadTraffic2::IPSet - Firewall handler for ipset lists

=head1 SYNOPSIS

This code is used internally.
  
=head1 DESCRIPTION

Example of a BadTraffic2 firewall handler.

=head1 INTERFACE 

There are no public methods.

=cut

=head2 new

Initialize the handler.

=cut

sub new {
    my ( $class, %parameters ) = @_;

    # defaults and valid keys
    my $self = bless(
        {   FW_NAME => 'badtraffic',
            DEBUG   => 0,
        },
        ref($class) || $class
    );

    foreach my $key ( sort keys %parameters ) {
        if ( defined $self->{$key} ) {
            $self->{$key} = $parameters{$key};
        }
        else {
            # quietly ignore extra params
        }
    }

    # test to make sure the handler execs can be used
    my $username = ( getpwuid $> );
    $self->{_IAMROOT} = ( $username eq 'root' ? 1 : 0 );

    # locate external executables
    if ( $self->{_IAMROOT} ) {
        $self->_find_execs();
        croak "could not find ipset executable" if ( !$self->{EXE_ipset} );
        croak "expected characters in set name $self->{FW_NAME}"
            if ( $self->{FW_NAME} !~ /\A[\w\-:]+\z/ );
        my $result;
        $result
            = `$self->{EXE_ipset} create $self->{FW_NAME} $SetType -exist 2>&1`;
        $result = `$self->{EXE_ipset} list $self->{FW_NAME} 2>&1`;
        croak $result if ( $result =~ /does not exist/ );
        print STDERR "BadTraffic2: Using firewall ipset $self->{FW_NAME}\n"
            if ( $self->{DEBUG} );
    }
    else {
        print STDERR "BadTraffic2: Not root, will not use ipset firewall\n"
            if ( $self->{DEBUG} );
    }

    # tell upstream code that we prefer to update individual entries
    $self->{_UPDATES} = 'entry';

    return ($self);
}

=head2 config_check 

Check and fix firewall configuration

=cut

sub config_check {
    my $self = shift;

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    # checks iptables rules for a matching block using the current set
    if ( !$self->{'EXE_iptables-save'} ) {
        carp "could not find executable: iptables-save";
        return;
    }
    my $result
        = `$self->{'EXE_iptables-save'} | grep -- '--match-set $self->{FW_NAME}'`;
    if ( !$result ) {
        if ( !$self->{EXE_iptables} ) {
            carp "could not find executable: iptables";
            return;
        }

        # write rules in same comment format used by puppet module
        system(   "$self->{EXE_iptables} -I INPUT "
                . "-m comment --comment '001 match-set badtraffic INPUT' "
                . "-m set --match-set badtraffic src -j DROP" );
        system(   "$self->{EXE_iptables} -I OUTPUT "
                . "-m comment --comment '001 match-set badtraffic OUTPUT' "
                . "-m set --match-set badtraffic src -j DROP" );
    }
}

=head2 read_set

Read list of blocked addresses as a set object.

=cut 

sub read_set {
    my $self = shift;
    my $raw  = shift;    # if true, return an array of entries

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    my $cidrset = Net::CIDR::Set->new( { type => 'ipv4' } );
    my $ct      = 0;
    my $rawlist = [];
    my $result;
    $result = `$self->{EXE_ipset} list $self->{FW_NAME} -s 2>&1`;
    foreach my $lin ( split /[\r\n]+/, $result ) {
        next if ( $lin =~ m{:\s} );                     # skip headers
        next if ( $lin !~ $BadTraffic2::CIDRregex );    # skip junk
        $cidrset->add($lin);
        push @$rawlist, $lin;
        $ct++;
    }
    print STDERR "ipset: read $ct entries\n" if ( $self->{DEBUG} );
    return ( $raw ? $rawlist : $cidrset );
}

=head2 write_set 

Write entire list of addresses, removing previous contents.

=cut

sub write_set {
    my $self    = shift;
    my $cidrset = shift;

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    # we won't usually do this, but...
    my $cur = $self->{FW_NAME};
    my $new = $cur . '.new';
    my $result;
    $result = `$self->{EXE_ipset} create $new $SetType -exist 2>&1`;
    $result = `$self->{EXE_ipset} flush $new 2>&1`;

    my $ct   = 0;
    my $iter = $cidrset->iterate_cidr;
    while ( my $entry = $iter->() ) {
        $entry .= '/32' if ( $entry !~ m{/} );    # fix single IPs to be /32s
        $result = `$self->{EXE_ipset} add $new $entry 2>&1`;
        $ct++;
    }
    print STDERR "ipset: write $ct entries\n" if ( $self->{DEBUG} );
    $result = `$self->{EXE_ipset} swap $new $cur 2>&1`;
    $result = `$self->{EXE_ipset} destroy $new 2>&1`;
}

=head2 add_entry 

Add a single ip/cidr rule to the existing list.

=cut

sub add_entry {
    my $self   = shift;
    my $ipcidr = shift;

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    if ( $ipcidr !~ $BadTraffic2::CIDRregex ) {
        carp "ipset: invalid entry $ipcidr";
        return;
    }

    # add it to the list, ignoring if it exists
    my $result;
    $result = `$self->{EXE_ipset} add $self->{FW_NAME} $ipcidr -exist 2>&1`;

    print STDERR "ipset: added $ipcidr\n" if ( $self->{DEBUG} );
}

=head2 del_entry

Remove a single ip/cidr rule from the existing list.

=cut 

sub del_entry {
    my $self   = shift;
    my $ipcidr = shift;

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    # ipcidr must be an exact match for a file entry
    $ipcidr .= '/32' if ( $ipcidr !~ m{/} );
    if ( $ipcidr !~ $BadTraffic2::CIDRregex ) {
        carp "ipset: invalid entry $ipcidr";
        return;
    }
    my $result;
    $result = `$self->{EXE_ipset} del $self->{FW_NAME} $ipcidr 2>&1`;
    if ( $result =~ m{not added} ) {
        print STDERR "ipset: remove $ipcidr not found\n"
            if ( $self->{DEBUG} );
    }
    else {
        print STDERR "ipset: removed $ipcidr\n" if ( $self->{DEBUG} );
    }
}

=head2 commit

Update synchronize all changes

=cut

sub commit {
    my $self = shift;
}

sub _find_execs {
    my $self = shift;

    my @PATHS = qw( /usr/sbin /sbin /usr/bin /usr/local/sbin /usr/local/bin );
    if ( my $env = $ENV{'PATH'} ) {
        @PATHS = ( split( /:/, $env ), @PATHS );
    }

    foreach my $cmd (qw( ipset iptables iptables-save )) {
        $self->{"EXE_$cmd"} = undef;
    PATHCHK:
        foreach my $path (@PATHS) {
            if ( -x "$path/$cmd" ) {
                $self->{"EXE_$cmd"} = "$path/$cmd";
                last PATHCHK;
            }
        }
        print STDERR qq(ipset: using executable $self->{"EXE_$cmd"}\n)
            if ( defined $self->{"EXE_$cmd"} && $self->{DEBUG} );
    }
}

1;    # Magic true value required at end of module
__END__

=head1 AUTHOR

Charles J. Brabec  C<< <brabec@ncsu.edu> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2019, Charles J. Brabec C<< <brabec@ncsu.edu> >>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
