package BadTraffic2::IPSetFwd;

use warnings;
use strict;
use Carp;

# exports
use base qw( Exporter );
our @EXPORT      = qw( );
our @EXPORT_OK   = qw( );
our %EXPORT_TAGS = (
    ALL => [ @EXPORT, @EXPORT_OK ],    # if :ALL requested
);

# this should match the settings used by the firewalld_ipset module in puppet
my $SetType = '--type=hash:net --family=inet '
    . '--option=hashsize=8192 --option=maxelem=131072';

=head1 NAME

BadTraffic2::IPSetFwd - Firewall handler for ipsets under Firewalld

=head1 SYNOPSIS

This code is used internally.
  
=head1 DESCRIPTION

Example of a BadTraffic2 firewall handler.

=head1 INTERFACE 

There are no public methods.

=cut

=head2 new

Initialize the handler.

=cut

sub new {
    my ( $class, %parameters ) = @_;

    # defaults and valid keys
    my $self = bless(
        {   FW_NAME => 'badtraffic',
            DEBUG   => 0,
        },
        ref($class) || $class
    );

    foreach my $key ( sort keys %parameters ) {
        if ( defined $self->{$key} ) {
            $self->{$key} = $parameters{$key};
        }
        else {
            # quietly ignore extra params
        }
    }

    # test to make sure the handler execs can be used
    my $username = ( getpwuid $> );
    $self->{_IAMROOT} = ( $username eq 'root' ? 1 : 0 );

    # locate external executables
    if ( $self->{_IAMROOT} ) {
        $self->_find_execs();
        croak "could not find firweall-cmd executable"
            if ( !$self->{EXEC} );
        croak "expected characters in set name $self->{FW_NAME}"
            if ( $self->{FW_NAME} !~ /\A[\w\-:]+\z/ );
        my $result;
        $result = `$self->{EXECP} --get-ipsets 2>&1`;
        $result = `$self->{EXECP} --new-ipset=$self->{FW_NAME} $SetType 2>&1`
            if ( $result !~ m{\b$self->{FW_NAME}\b} );
        $result = `$self->{EXECP} --info-ipset=$self->{FW_NAME} 2>&1`;
        croak $result if ( $result =~ /Error:/ );
        print STDERR "BadTraffic2: Using firewall ipset $self->{FW_NAME}\n"
            if ( $self->{DEBUG} );
    }
    else {
        print STDERR "BadTraffic2: Not root, will not use ipset firewall\n"
            if ( $self->{DEBUG} );
    }

    # tell upstream code that we prefer to update by rewriting the list
    $self->{_UPDATES} = 'all';

    return ($self);
}

=head2 config_check 

Check and fix firewall configuration

=cut

sub config_check {
    my $self = shift;

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    # checks iptables rules for a matching block using the current set
    if ( !$self->{'EXEC'} ) {
        carp "could not find executable: firewall-cmd";
        return;
    }
    my $result
        = `$self->{EXECP} --list-rich-rules | grep 'ipset=.$self->{FW_NAME}'`;
    if ( !$result ) {

        # write rule in same format used by puppet module
        system(   "$self->{EXECP} --add-rich-rule "
                . qq('rule family="ipv4" source ipset="$self->{FW_NAME}" drop')
        );
        $self->commit();    # reload firewalld after change
    }
}

=head2 read_set

Read list of blocked addresses as a set object.

=cut 

sub read_set {
    my $self = shift;
    my $raw  = shift;       # if true, return an array of entries

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    my $cidrset = Net::CIDR::Set->new( { type => 'ipv4' } );
    my $ct      = 0;
    my $rawlist = [];
    my $result;
    $result = `$self->{EXECP} --ipset=$self->{FW_NAME} --get-entries 2>&1`;
    foreach my $lin ( split /[\r\n]+/, $result ) {
        next if ( $lin =~ m{:\s} );                     # skip headers
        next if ( $lin !~ $BadTraffic2::CIDRregex );    # skip junk
        $cidrset->add($lin);
        push @$rawlist, $lin;
        $ct++;
    }
    print STDERR "ipset: read $ct entries\n" if ( $self->{DEBUG} );
    return ( $raw ? $rawlist : $cidrset );
}

=head2 write_set 

Write entire list of addresses, removing previous contents.

=cut

sub write_set {
    my $self    = shift;
    my $cidrset = shift;

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    # firewalld doesn't have a function for purge or swap
    # going to grab the permanent config file and just rewrite it
    my $xmlfile;
    $xmlfile = `$self->{EXECP} --path-ipset=$self->{FW_NAME}`;
    chomp($xmlfile);
    croak "Unable to find firewalld ipset config file"
        if ( !-s $xmlfile );
    my $newfile = $xmlfile . '.new';
    open( my $in,  '<', $xmlfile ) || croak "Failed to read $xmlfile: $!\n";
    open( my $out, '>', $newfile ) || croak "Failed to write $newfile: $!\n";

    # copy xmlfile to newfile until the entries start
    while ( my $lin = <$in> ) {
        last if ( $lin =~ m{<entry>|</ipset>} );
        print $out $lin;
    }

    # fill new newfile with new entries
    my $ct   = 0;
    my $iter = $cidrset->iterate_cidr;
    while ( my $entry = $iter->() ) {
        $entry .= '/32' if ( $entry !~ m{/} );    # fix single IPs to be /32s
        print $out "  <entry>$entry</entry>\n";
        $ct++;
    }

    # now find the rest of xmlfile can finish it
    while ( my $lin = <$in> ) {
        next if ( $lin =~ m{<entry>} );
        last if ( $lin =~ m{</ipset>} );          # don't pass final close tag
        print $out $lin;
    }

    # explicit final close tag (fixes bug with no old entries)
    print $out "</ipset>\n";
    close($in);
    close($out);

    # swap the files
    system 'mv', $xmlfile, "$xmlfile.old";
    system 'mv', $newfile, $xmlfile;

    print STDERR "ipset: write $ct entries\n" if ( $self->{DEBUG} );
    $self->commit();
}

=head2 add_entry 

Add a single ip/cidr rule to the existing list.

=cut

sub add_entry {
    my $self   = shift;
    my $ipcidr = shift;

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    if ( $ipcidr !~ $BadTraffic2::CIDRregex ) {
        carp "ipset: invalid entry $ipcidr";
        return;
    }

    # add it to the list, ignoring if it exists
    my $result;
    $result
        = `$self->{EXECP} --ipset=$self->{FW_NAME} --add-entry=$ipcidr 2>&1`;

    print STDERR "ipset: added $ipcidr\n" if ( $self->{DEBUG} );
}

=head2 del_entry

Remove a single ip/cidr rule from the existing list.

=cut 

sub del_entry {
    my $self   = shift;
    my $ipcidr = shift;

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    # ipcidr must be an exact match for a file entry
    $ipcidr .= '/32' if ( $ipcidr !~ m{/} );
    if ( $ipcidr !~ $BadTraffic2::CIDRregex ) {
        carp "ipset: invalid entry $ipcidr";
        return;
    }
    my $result;
    $result
        = `$self->{EXECP} --ipset=$self->{FW_NAME} --remove-entry=$ipcidr 2>&1`;
    if ( $result =~ m{not_enabled}i ) {
        print STDERR "ipset: remove $ipcidr not found\n"
            if ( $self->{DEBUG} );
    }
    else {
        print STDERR "ipset: removed $ipcidr\n" if ( $self->{DEBUG} );
    }
}

=head2 commit

Update synchronize all changes

=cut 

sub commit {
    my $self = shift;

    croak "FW_TYPE ipset must be run as root"
        if ( !$self->{_IAMROOT} );

    # checks iptables rules for a matching block using the current set
    if ( !$self->{'EXEC'} ) {
        carp "could not find executable: firewall-cmd";
        return;
    }
    my $result = `$self->{EXEC} --reload 2>&1`;
    carp($result) if ( $result !~ /success/ );
}

sub _find_execs {
    my $self = shift;

    my @PATHS = qw( /usr/sbin /sbin /usr/bin /usr/local/sbin /usr/local/bin );
    if ( my $env = $ENV{'PATH'} ) {
        @PATHS = ( split( /:/, $env ), @PATHS );
    }

    foreach my $cmd (qw( firewall-cmd )) {
        my $label = "EXEC";
        $self->{$label} = undef;
    PATHCHK:
        foreach my $path (@PATHS) {
            if ( -x "$path/$cmd" ) {
                $self->{$label} = "$path/$cmd";
                last PATHCHK;
            }
        }
        print STDERR qq(ipset: using executable $self->{$label}\n)
            if ( defined $self->{$label} && $self->{DEBUG} );
    }
    $self->{EXECP} = "$self->{EXEC} --permanent";
}

1;    # Magic true value required at end of module
__END__

=head1 AUTHOR

Charles J. Brabec  C<< <brabec@ncsu.edu> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2019, Charles J. Brabec C<< <brabec@ncsu.edu> >>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
