package BadTraffic2::SDBM;

use warnings;
use strict;
use Carp;
use Fcntl;
use File::Copy qw(move);
use SDBM_File;

# exports
use base qw( Exporter );
our @EXPORT      = qw( );
our @EXPORT_OK   = qw( );
our %EXPORT_TAGS = (
    ALL => [ @EXPORT, @EXPORT_OK ],    # if :ALL requested
);

=head1 NAME

BadTraffic2::SDBM - Firewall handler for SDBM files

=head1 SYNOPSIS

This code is used internally.
  
=head1 DESCRIPTION

Example of a BadTraffic2 firewall handler.

=head1 INTERFACE 

There are no public methods.

=cut

=head2 new

Initialize the handler.

=cut

sub new {
    my ( $class, %parameters ) = @_;

    # defaults and valid keys
    my $self = bless(
        {   FW_NAME => '/tmp/badtraffic_sdbm',
            EXPAND  => 28,
            DEBUG   => 0,
        },
        ref($class) || $class
    );

    foreach my $key ( sort keys %parameters ) {
        if ( defined $self->{$key} ) {
            $self->{$key} = $parameters{$key};
        }
        else {
            # quietly ignore extra params
        }
    }

    # tests to make sure the handler can be used
    croak "FW_TYPE sdbm requires FW_NAME be set to a file path"
        if ( !$self->{FW_NAME} );
    print STDERR "BadTraffic2: Using firewall sdbm $self->{FW_NAME}\n"
        if ( $self->{DEBUG} );

    # tell upstream code that we prefer to rewrite the entire list
    $self->{_UPDATES} = 'all';

    return ($self);
}

=head2 config_check 

Check and fix firewall configuration

=cut

sub config_check {
    my $self = shift;

}

=head2 read_set

Read list of blocked addresses as a set object.

=cut 

sub read_set {
    my $self = shift;
    my $raw  = shift;    # if true, return an array of entries

    my $cidrset = Net::CIDR::Set->new( { type => 'ipv4' } );
    my $ct      = 0;
    my $rawlist = [];
    if ( my $in = $self->_open_tie() ) {
        while ( my ( $ip, $block ) = each(%$in) ) {
            next if ( $block != 1 );    # skip if not blocked
            if ( $ip =~ $BadTraffic2::CIDRregex ) {
                $cidrset->add($ip);
                push @$rawlist, $ip;
                $ct++;
            }
            else {
                carp "Skipped bad CIDR: $ip" if ( $self->{DEBUG} );
            }
        }
        $self->_close_tie();
    }
    elsif ( -s $self->{FW_NAME} ) {
        croak "sdbm $self->{FW_NAME} is not readable";
    }
    print STDERR "sdbm: read $ct entries\n" if ( $self->{DEBUG} );
    return ( $raw ? $rawlist : $cidrset );
}

=head2 write_set 

Write entire list of addresses, removing previous contents.

=cut

sub write_set {
    my $self    = shift;
    my $cidrset = shift;

    my $ct  = 0;
    my $cur = $self->{FW_NAME};
    my $new = $cur . '.new';
    my $out = $self->_open_tie($new)
        || croak "Cannot write new sdbm files: $!";
    my $iter = $cidrset->iterate_cidr;
    while ( my $ipcidr = $iter->() ) {
        if ( $ipcidr !~ m{/} ) {
            $out->{$ipcidr} = 1;
            $ct++;
        }
        else {
            my ( $address, $bits ) = split m{/}, $ipcidr, 2;
            if ( $bits >= $self->{EXPAND} ) {
                my $subset = Net::CIDR::Set->new($ipcidr);
                my $adds   = $subset->iterate_addresses;
                while ( my $ip = $adds->() ) {
                    $out->{$ip} = 1;
                    $ct++;
                }
                $out->{$ipcidr} = 1;    # save cidr as sync marker
            }
            else {
                carp "Skipped large block: $ipcidr" if ( $self->{DEBUG} );
            }
        }
    }
    $self->_close_tie();
    print STDERR "sdbm: write $ct entries\n" if ( $self->{DEBUG} );

    # have to replace both extensions
    foreach my $ext ( 'dir', 'pag' ) {
        move( "$new.$ext", "$cur.$ext" )
            || croak "Failed to replace sdbm .$ext file";
    }
}

=head2 add_entry 

Add a single ip/cidr rule to the existing list.

=cut

sub add_entry {
    my $self   = shift;
    my $ipcidr = shift;

    if ( $ipcidr !~ $BadTraffic2::CIDRregex ) {
        carp "sdbm: invalid entry $ipcidr";
        return;
    }

    # add it to the list, ignoring if it exists
    my $ct  = 0;
    my $out = $self->_open_tie()
        || croak "Cannot write new sdbm files: $!";
    if ( $ipcidr !~ m{/} ) {
        $out->{$ipcidr} = 1;
        $ct++;
    }
    else {
        my ( $address, $bits ) = split m{/}, $ipcidr, 2;
        if ( $bits >= $self->{EXPAND} ) {
            my $subset = Net::CIDR::Set->new($ipcidr);
            my $adds   = $subset->iterate_addresses;
            while ( my $ip = $adds->() ) {
                $out->{$ip} = 1;
                $ct++;
            }
            $out->{$ipcidr} = 1;    # save cidr as sync marker
        }
        else {
            carp "Skipped large block: $ipcidr" if ( $self->{DEBUG} );
            return;
        }
    }
    print STDERR "sdbm: added $ct IPs for $ipcidr\n" if ( $self->{DEBUG} );
}

=head2 del_entry

Remove a single ip/cidr rule from the existing list.

=cut 

sub del_entry {
    my $self   = shift;
    my $ipcidr = shift;
    if ( $ipcidr !~ $BadTraffic2::CIDRregex ) {
        carp "sdbm: invalid entry $ipcidr";
        return;
    }

    # remove it from the list, ignoring if it exists
    my $ct  = 0;
    my $out = $self->_open_tie()
        || croak "Cannot write new sdbm files: $!";
    if ( $ipcidr !~ m{/} ) {
        delete $out->{$ipcidr};
        $ct++;
    }
    else {
        my ( $address, $bits ) = split m{/}, $ipcidr, 2;
        if ( $bits >= $self->{EXPAND} ) {
            my $subset = Net::CIDR::Set->new($ipcidr);
            my $adds   = $subset->iterate_addresses;
            while ( my $ip = $adds->() ) {
                delete $out->{$ip};
                $ct++;
            }
            delete $out->{$ipcidr};    # remove cidr sync marker
        }
        else {
            carp "Skipped large block: $ipcidr" if ( $self->{DEBUG} );
            return;
        }
    }
    print STDERR "sdbm: removed $ct IPs for $ipcidr\n" if ( $self->{DEBUG} );
}

=head2 commit

Update synchronize all changes

=cut

sub commit {
    my $self = shift;
}

=head2 _open_tie

Returns a tied hash to the SDBM file.

=cut

sub _open_tie {
    my $self = shift;
    my $filename = shift || $self->{FW_NAME};

    return $self->{_TIE} if ( defined $self->{_TIE} );

    my $href;
    if ( tie( %$href, 'SDBM_File', $filename, O_RDWR|O_CREAT, oct(666) ) ) {
        $self->{_TIE} = $href;
        return $href;
    }

    carp("Failed to open SDBM $filename: $!") if ( $self->{DEBUG} );
    delete $self->{_TIE};
    return;
}

=head2 _close_tie

Removes tied hash from object.

=cut

sub _close_tie {
    my $self = shift;
    untie( %{ $self->{_TIE} } ) if ( defined $self->{_TIE} );
    delete $self->{_TIE};
    return;
}

1;    # Magic true value required at end of module
__END__

=head1 AUTHOR

Charles J. Brabec  C<< <brabec@ncsu.edu> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2019, Charles J. Brabec C<< <brabec@ncsu.edu> >>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
