package BadTraffic2;

use warnings;
use strict;
use Carp;
use JSON qw();
use NCSU::RESTclient;
use Net::CIDR::Set;
use Sys::Hostname qw(hostname);
use URI;

# our submodules
use BadTraffic2::IPSet;
use BadTraffic2::IPSetFwd;
use BadTraffic2::FlatFile;
use BadTraffic2::SDBM;
use BadTraffic2::Template;    # empty template used for type = 'none'

use version; our $VERSION = qv('2.3.2');

# exports
use base qw( Exporter );
our @EXPORT      = qw( );
our @EXPORT_OK   = qw( );
our %EXPORT_TAGS = (
    ALL => [ @EXPORT, @EXPORT_OK ],    # if :ALL requested
);

# Common regexes
our $CIDRregex
    = qr{\A(?:[0-9]{1,3}\.){3}[0-9]{1,3}(?:/(?:[0-9]|[1-2][0-9]|3[0-2]))?\z};

=head1 NAME

BadTraffic2 - Perl extension for IP reputation firewall management


=head1 VERSION

This document describes BadTraffic2 version 2.1.0


=head1 SYNOPSIS

    use BadTraffic2;
    
    # load the module
    my $bt = BadTraffic2->new(
        LIST    => 'web-all',
        FW_TYPE => 'ipset',
        FW_NAME => 'badtraffic',
    );


    # sync the chosen firewall with the lists from the database
    # with addition of local blocks, if found
    my $blockset = $bt->get_db_blocks();
    my $localset = $bt->get_local_blocks();
    $blockset->add($localset);
    $bt->put_fw_blocks($blockset);

    # add a rule to the database
    my $rule = {
        cidr        => '5.1.2.3/32',
        listname    => 'web-cp',
        reason      => 'scanning wordpress for exploits',
        duration    => '90d',
    };
    my $result = $bt->add_rule( $rule );
    die "Add rule failed: $result->{ERROR}\n" if ($result->{ERROR});
    print "Added new rule $result->{id}\n";

  
=head1 DESCRIPTION

BadTraffic2 reads and writes to a MySQL database and blocks traffic
according entries found. It provides facilities for allowlisting
addresses or networks using CIDR formatting.


=head1 INTERFACE 

Each public function/method is described here.
These are how you should interact with this module.

=cut

=head2 new

 Usage    : my $bt = new BadTraffic2;
 Purpose  : Create a new BadTraffic object.
 Returns  : An object in the form of a blessed hash ref.
 Argument : hash of values to set:
          :    CONFIG      => config file (default: /etc/badtraffic)
          :    LOCAL_FILE  => file containing locally blocked cidrs
          :    REST_ID     => REST interface key ID for writes
          :    REST_KEY    => REST interface secret key
          :    REST_BASE   => default: https://web01rest.unity.ncsu.edu/badtraffic
          :    FW_TYPE     => type of firewall (default: ipset)
          :    FW_PROVIDER => ipset only, which provider is implementing the ipset
          :                   iptables or firewalld (def: iptables)
          :    FW_NAME     => name / filepath of FW data (def: badtraffic)
          :    LIST        => list to subscribe (default: all)
          :    EXPAND      => sdbm only, CIDR size to expand into 
          :                   list of IP addresses (default: 28)
          :    DEBUG       => debug level (default: 0)
 Comments : Parameters passed will override config file settings.
          : The valid firewall types are "ipset", "sdbm", and "flatfile",
          :     or use "none" to bypass firewall functions.
          : Use FW_PROVIDER to select iptables using kernel ipsets 
          :     vs firewalld's internal ipsets.
          : For 'ipset', the FW_NAME is the name of the ipset.
          : For 'sdbm', the FW_NAME is the full path of the file, 
          :     not including the .pag or .dir extensions.
          : For 'flatfile', the FW_NAME is the full path of the file.
          : The REST_HOST is required for all connections.
          : The REST ID/KEY is only required to perform updates to the
          :     shared database.

=cut

sub new {
    my ( $class, %parameters ) = @_;

    # defaults and valid keys
    my $self = bless(
        {   CONFIG      => '/etc/badtraffic.conf',
            LOCAL_FILE  => '',
            REST_ID     => '',
            REST_KEY    => '',
            REST_BASE   => 'https://web01rest.unity.ncsu.edu/badtraffic',
            FW_TYPE     => 'ipset',
            FW_NAME     => 'badtraffic',
            FW_PROVIDER => 'iptables',
            LIST        => 'undef',
            EXPAND      => 28,
            DEBUG       => 0,
        },
        ref($class) || $class
    );

    # CONFIG can be set in env
    $self->{CONFIG} = $ENV{'BT_CONFIG'}
        if ( defined $ENV{'BT_CONFIG'} && -e $ENV{'BT_CONFIG'} );

    # look for CONFIG setting before checking config file
    $self->{CONFIG} = $parameters{CONFIG} if ( defined $parameters{CONFIG} );
    $self->_read_config();

    # override with any passed paramteres
    foreach my $key ( sort keys %parameters ) {
        if ( defined $self->{$key} ) {
            $self->{$key} = $parameters{$key};
        }
        else {
            carp "new: Ignored unknown parameter $key";
        }
    }

    # initialize the firewall handler
    $self->{FW_HANDLER} = (
        $self->{FW_TYPE} eq 'ipset'
        ? ( $self->{FW_PROVIDER} eq 'firewalld'
            ? BadTraffic2::IPSetFwd->new(%$self)
            : BadTraffic2::IPSet->new(%$self)
            )
        : $self->{FW_TYPE} eq 'flatfile' ? BadTraffic2::FlatFile->new(%$self)
        : $self->{FW_TYPE} eq 'sdbm'     ? BadTraffic2::SDBM->new(%$self)
        : $self->{FW_TYPE} eq 'none'     ? BadTraffic2::Template->new(%$self)
        :   croak "new: unknown FW_TYPE = $self->{FW_TYPE}"
    );

    # initialize the REST handler
    my $client = NCSU::RESTclient->new(
        {   baseurl => $self->{REST_BASE},
            encoder => 'json',               # server side requires json input
        }
    );
    $client->set_auth(
        'statickey',
        {   keyid   => $self->{REST_ID},
            keydata => $self->{REST_KEY},
        }
    );
    $self->{REST_CLIENT} = $client;

    return ($self);
}

=head2 fw_config_check 

 Usage    : $bt->fw_config_check();
 Purpose  : Check and enable the firewall rules in iptables.
 Returns  : Nothing
 Comments : Only useful for 'ipset' firewalls.
          : This function should be handled externally by puppet.
          : This code will try to write the same config as puppet.
 
=cut 

sub fw_config_check {
    my $self = shift;
    my $result;

    if ( defined $self->{FW_HANDLER} ) {
        $result = $self->{FW_HANDLER}->config_check();
    }
    else {
        carp "fw_config_check: no firewall handler";
    }
    return $result;
}

=head2 get_fw_blocks

 Usage    : $bt->get_fw_blocks();
 Purpose  : Get set of active blocks on the firewall.
 Returns  : A Set object containing the blocked CIDR/IP addresses
 
=cut

sub get_fw_blocks {
    my $self    = shift;
    my $cidrset = Net::CIDR::Set->new( { type => 'ipv4' } );

    if ( defined $self->{FW_HANDLER} ) {
        $cidrset = $self->{FW_HANDLER}->read_set();
    }
    else {
        carp "get_fw_blocks: no firewall handler";
    }
    return $cidrset;
}

=head2 put_fw_blocks

 Usage    : $bt->put_fw_blocks($cidrset);
 Purpose  : Write the given set of blocks to the firewall.
 Argument : $cidrset = a Net::CIDR::Set of blocked addresses
 Returns  : Nothing.
 
=cut

sub put_fw_blocks {
    my $self    = shift;
    my $cidrset = shift;
    if ( !UNIVERSAL::isa( $cidrset, 'Net::CIDR::Set' ) ) {
        carp "put_fw_blocks: did not get a Set object";
        return;
    }
    if ( !defined $self->{FW_HANDLER} ) {
        carp "put_fw_blocks: no firewall handler";
        return;
    }

    if ( $self->{FW_HANDLER}->{_UPDATES} eq 'entry' ) {

        # process differences between new list and old
        my $fwlist  = $self->{FW_HANDLER}->read_set('raw');
        my $fwhash  = { map { $_ => 1 } @$fwlist };
        my $newlist = [ $cidrset->as_cidr_array ];
        my $newhash = { map { $_ => 1 } @$newlist };

        # deletes first
        foreach my $entry (@$fwlist) {
            next if ( $newhash->{$entry} );
            $self->{FW_HANDLER}->del_entry($entry);
        }

        # adds next
        foreach my $entry (@$newlist) {
            next if ( $fwhash->{$entry} );
            $self->{FW_HANDLER}->add_entry($entry);
        }

        # commit changes for types that need it
        $self->{FW_HANDLER}->commit();
    }
    else {

        # process as a single write
        $self->{FW_HANDLER}->write_set($cidrset);

    }
}

=head2 add_fw_block

 Usage    : $bt->add_fw_block($cidr);
 Purpose  : Adds a single CIDR block to the firewall.
 Argument : $cidr = an IP or CIDR string to block
 Returns  : Nothing.

=cut

sub add_fw_block {
    my $self = shift;
    my $cidr = shift;

    if ( !defined $self->{FW_HANDLER} ) {
        carp "add_fw_block: no firewall handler";
        return;
    }
    if ( UNIVERSAL::isa( $cidr, 'Net::CIDR::Set' ) ) {

        # might be more than one, just go with it
        my $iter = $cidr->iterate_cidr;
        while ( my $entry = $iter->() ) {
            $entry .= '/32' if ( $entry !~ m{/} ); # fix single IPs to be /32s
            $self->{FW_HANDLER}->add_entry($entry);
        }
    }
    elsif ( $cidr =~ $CIDRregex ) {
        $cidr .= '/32' if ( $cidr !~ m{/} );       # fix single IPs to be /32s
        $self->{FW_HANDLER}->add_entry($cidr);
    }
    else {
        carp "add_fw_block: unrecognized block '$cidr'";
    }

    # commit changes for types that need it
    $self->{FW_HANDLER}->commit();
}

=head2 del_fw_block

 Usage    : $bt->del_fw_block($cidr);
 Purpose  : Remove a single CIDR block from the firewall.
 Argument : $cidr = an IP or CIDR string to remove
 Returns  : Nothing.
 Comments : $cidr must be an exact match for a rule in the set.
          : This will not remove IPs from inside a larger range.
          : Use Net::CIDR::Set and put_fw_blocks to do deep merging.

=cut

sub del_fw_block {
    my $self = shift;
    my $cidr = shift;

    if ( !defined $self->{FW_HANDLER} ) {
        carp "del_fw_block: no firewall handler";
        return;
    }
    if ( UNIVERSAL::isa( $cidr, 'Net::CIDR::Set' ) ) {

        # might be more than one, just go with it
        my $iter = $cidr->iterate_cidr;
        while ( my $entry = $iter->() ) {
            $self->{FW_HANDLER}->del_entry($entry);
        }
    }
    elsif ( $cidr =~ $CIDRregex ) {
        $self->{FW_HANDLER}->del_entry($cidr);
    }
    else {
        carp "del_fw_block: unrecognized block '$cidr'";
    }

    # commit changes for types that need it
    $self->{FW_HANDLER}->commit();
}

=head2 get_db_blocks

 Usage    : my $blockset = $bt->get_db_blocks($listname);
 Purpose  : Get current set of active blocks from the database 
 Argument : A listname, or uses LIST set in the object.
 Returns  : A CIDR set containing the blocked CIDR/IP addresses
          : On error, carps and returns undef.
 
=cut

sub get_db_blocks {
    my $self     = shift;
    my $listname = shift;
    return $self->_get_db_static( 'bl', $listname );
}

=head2 get_db_allows

 Usage    : my $allowset = $bt->get_db_allows($listname);
 Purpose  : Get current set of allowlisted entries from the database 
 Argument : A listname, or uses LIST set in the object.
 Returns  : A CIDR set containing the allowlisted CIDR/IP addresses
          : On error, carps and returns undef.
 
=cut

sub get_db_allows {
    my $self     = shift;
    my $listname = shift;
    return $self->_get_db_static( 'wl', $listname );
}

=head2 get_db_whites

 Deprecated. Change code to use get_db_allows().

=cut

sub get_db_whites { return get_db_allows(@_); }

=head2 get_db_recent

 Usage    : my $blockct = $bt->get_db_recent($days);
 Purpose  : Get count of recent but expired blocks, per cidr.
 Argument : Days passed to recent, should be one of [30,45,60,90]
 Returns  : A hashref of cidr => block_count
          : On error, carps and returns undef.
 
=cut

sub get_db_recent {
    my $self = shift;
    my $days = shift || 30;

    # convert requested days to a valid interval
    my @intervals = ( 30, 45, 60, 90 );
    foreach my $int (@intervals) {
        if ( $days <= $int ) {
            $days = $int;
            last;
        }
    }
    $days = $intervals[-1] if ( $days > $intervals[-1] );

    return $self->_get_db_static( $days, 'recent' );
}

=head2 get_local_blocks

 Usage    : $bt->get_local_blocks();
 Purpose  : Get set of local blocks from LOCAL_FILE.
 Returns  : A Set object containing the blocked CIDR/IP addresses
 
=cut

sub get_local_blocks {
    my $self    = shift;
    my $cidrset = Net::CIDR::Set->new( { type => 'ipv4' } );

    # semi-silent return
    if ( !$self->{LOCAL_FILE} || !-r $self->{LOCAL_FILE} ) {
        carp('get_local_blocks: LOCAL_FILE not found') if ( $self->{DEBUG} );
        return $cidrset;
    }

    # use the flatfile handler to read the local file
    my $handler
        = BadTraffic2::FlatFile->new( FW_NAME => $self->{LOCAL_FILE}, );
    $cidrset = $handler->read_set;
    my $ct = $cidrset->as_cidr_array;
    print STDERR "get_local_blocks: read $ct entries\n" if ( $self->{DEBUG} );
    return $cidrset;
}

=head2 get_blocked_status

 Usage    : my $result = $bt->get_blocked_status($cidr);
 Purpose  : Get blocked status report for given CIDR address.
 Argument : A CIDR address
 Returns  : A hashref containing the status data.
          : { ERROR => $msg } on any errors.
 Comments : For return details, see REST document for GET /blocked/:cidr

=cut

sub get_blocked_status {
    my $self = shift;
    my $cidr = shift;

    # might be more than one, take the first one only
    if ( UNIVERSAL::isa( $cidr, 'Net::CIDR::Set' ) ) {
        my $iter = $cidr->iterate_cidr;
        $cidr = $iter->();
    }

    return ( { ERROR => "unrecognized CIDR address '$cidr'" } )
        if ( $cidr !~ $CIDRregex );

    my $url = '/blocked/' . $cidr;
    my ( $result, $code, $msg ) = $self->_make_rest_call( 'GET' => $url );

    # success
    return $result if ( $code < 300 );

    # bad result
    return ( { ERROR => $msg } );
}

=head2 search_rules 

 Usage    : my $result = $bt->search_rules($params);
 Purpose  : Calls GET /rules to search the rules table
 Argument : A hashref of search params and values
 Returns  : An arrayref containing the list of matching rule hashes.
          : [ ERROR, $msg ] on any errors.
 Comment  : Refer to the REST doc for usage of search params.

=cut

sub search_rules {
    my $self   = shift;
    my $params = shift || {};
    $params = { $params, @_ } if ( ref $params ne 'HASH' );   # coerce hashref

    # need to make a query from the search params
    my $uriobj = URI->new('http:');
    $uriobj->query_form($params);
    my $query = $uriobj->query;
    return ( [ ERROR => 'no search params' ] ) if ( !$query );
    print STDERR "search_rules: prepared query $query\n"
        if ( $self->{DEBUG} );

    my $url = '/rules?' . $query;
    my ( $result, $code, $msg ) = $self->_make_rest_call( 'GET' => $url );

    # success
    return $result if ($result);

    # bad result
    return ( [ ERROR => $msg ] );
}

=head2 get_rule 

 Usage    : my $result = $bt->get_rule($id);
 Purpose  : Calls GET /rule/:id
 Argument : A rule id number
 Returns  : A hashref containing the rule data.
          : { ERROR => $msg } on any errors.

=cut

sub get_rule {
    my $self = shift;
    my $id   = shift;

    return ( { ERROR => "id='$id' must be a bare number" } )
        if ( $id !~ /\A\d+\z/ );

    my $url = '/rule/' . $id;
    my ( $result, $code, $msg ) = $self->_make_rest_call( 'GET' => $url );

    # success
    return $result if ($result);

    # bad result
    return ( { ERROR => $msg } );
}

=head2 add_rule 

 Usage    : my $result = $bt->add_rule($rulehash);
 Purpose  : Calls POST /rule
 Argument : A hashref containing rule data
 Returns  : A hashref containing the saved rule data.
          : { ERROR => $msg } on any errors.

=cut

sub add_rule {
    my $self = shift;
    my $href = shift || {};
    $href = { $href, @_ } if ( ref $href ne 'HASH' );    # coerce hashref

    # sneak this in if not passed
    $href->{reported_by} = $self->_default_reported_by()
        if ( !defined $href->{reported_by} );

    my $url = '/rule';
    my ( $result, $code, $msg )
        = $self->_make_rest_call( 'POST' => $url, $href );

    # success
    return $result if ($result);

    # bad result
    return ( { ERROR => $msg } );
}

=head2 edit_rule 

 Usage    : my $result = $bt->editrule($rulehash);
 Purpose  : Calls PUT /rule/:id
 Argument : A hashref containing rule data
 Returns  : A hashref containing the updated rule data.
          : { ERROR => $msg } on any errors.

=cut

sub edit_rule {
    my $self = shift;
    my $href = shift || {};
    $href = { $href, @_ } if ( ref $href ne 'HASH' );    # coerce hashref

    my $id = $href->{id} || '';
    return ( { ERROR => "id='$id' must be a bare number" } )
        if ( $id !~ /\A\d+\z/ );
    my $url = '/rule/' . $id;
    my ( $result, $code, $msg )
        = $self->_make_rest_call( 'PUT' => $url, $href );

    # success
    return $result if ($result);

    # bad result
    return ( { ERROR => $msg } );
}

=head2 get_geoip

 Usage    : my $result = $bt->get_geoip($cidr);
 Purpose  : Get GeoIP location information for an address
 Argument : A CIDR address
 Returns  : A hashref containing the geoip data.
          : { ERROR => $msg } on any errors.
 Comments : For return details, see REST document for GET /geoip/:cidr

=cut

sub get_geoip {
    my $self = shift;
    my $cidr = shift;

    # might be more than one, take the first one only
    if ( UNIVERSAL::isa( $cidr, 'Net::CIDR::Set' ) ) {
        my $iter = $cidr->iterate_cidr;
        $cidr = $iter->();
    }

    return ( { ERROR => "unrecognized CIDR address '$cidr'" } )
        if ( $cidr !~ $CIDRregex );

    my $url = '/geoip/' . $cidr;
    my ( $result, $code, $msg ) = $self->_make_rest_call( 'GET' => $url );

    # success
    return $result if ( $code < 300 );

    # bad result
    return ( { ERROR => $msg } );
}

=head2 get_all_lists  

 Usage    : my $result = $bt->get_all_lists();
 Purpose  : Calls GET /lists
 Returns  : A hashref containing {listname => {list data}, ... }
          : { ERROR => $msg } on any errors.

=cut

sub get_all_lists {
    my $self = shift;

    my $url = '/lists';
    my ( $result, $code, $msg ) = $self->_make_rest_call( 'GET' => $url );

    # success
    return $result if ($result);

    # bad result
    return ( { ERROR => $msg } );
}

=head2 get_list 

 Usage    : my $result = $bt->get_list($listname);
 Purpose  : Calls GET /list/:listname
 Argument : A listname, or use $self->{LIST}
 Returns  : A hashref containing the list data.
          : { ERROR => $msg } on any errors.

=cut

sub get_list {
    my $self     = shift;
    my $listname = shift || $self->{LIST};

    return (
        { ERROR => "listname '$listname' contains unexpected characters" } )
        if ( $listname !~ /\A[\w\-]+\z/ );

    my $url = '/list/' . $listname;
    my ( $result, $code, $msg ) = $self->_make_rest_call( 'GET' => $url );

    # success
    return $result if ($result);

    # bad result
    return ( { ERROR => $msg } );
}

=head2 add_list 

 Usage    : my $result = $bt->add_list($listhash);
 Purpose  : Calls POST /list
 Argument : A hashref containing listname and descript 
 Returns  : A hashref containing the saved list data.
          : { ERROR => $msg } on any errors.

=cut

sub add_list {
    my $self = shift;
    my $href = shift || {};
    $href = { $href, @_ } if ( ref $href ne 'HASH' );    # coerce hashref

    my $listname = $href->{listname} || '';
    return (
        { ERROR => "listname '$listname' contains unexpected characters" } )
        if ( $listname !~ /\A[\w\-]+\z/ );

    my $url = '/list';
    my ( $result, $code, $msg )
        = $self->_make_rest_call( 'POST' => $url, $href );

    # success
    return $result if ($result);

    # bad result
    return ( { ERROR => $msg } );
}

=head2 edit_list 

 Usage    : my $result = $bt->edit_list($listhash);
 Purpose  : Calls PUT /list/:listname
 Argument : A hashref containing listname and descript 
 Returns  : A hashref containing the updated list data.
          : { ERROR => $msg } on any errors.

=cut

sub edit_list {
    my $self = shift;
    my $href = shift || {};
    $href = { $href, @_ } if ( ref $href ne 'HASH' );    # coerce hashref

    my $listname = $href->{listname} || '';
    return (
        { ERROR => "listname '$listname' contains unexpected characters" } )
        if ( $listname !~ /\A[\w\-]+\z/ );

    my $url = '/list/' . $listname;
    my ( $result, $code, $msg )
        = $self->_make_rest_call( 'PUT' => $url, $href );

    # success
    return $result if ($result);

    # bad result
    return ( { ERROR => $msg } );
}

=head2 get_all_includes  

 Usage    : my $result = $bt->get_all_includes();
 Purpose  : Calls GET /includes
 Returns  : A hashref containing {listname => [ includes ], ... }
          : { ERROR => $msg } on any errors.

=cut

sub get_all_includes {
    my $self = shift;

    my $url = '/includes';
    my ( $result, $code, $msg ) = $self->_make_rest_call( 'GET' => $url );

    # success
    return $result if ($result);

    # bad result
    return ( { ERROR => $msg } );
}

=head2 get_includes 

 Usage    : my $result = $bt->get_includes($listname);
 Purpose  : Calls GET /includes/:listname
 Argument : A listname, or use $self->{LIST}
 Returns  : An arrayref containing the list of included listnames
          : [ ERROR, $msg ] on any errors.

=cut

sub get_includes {
    my $self     = shift;
    my $listname = shift || $self->{LIST};

    return (
        { ERROR => "listname '$listname' contains unexpected characters" } )
        if ( $listname !~ /\A[\w\-]+\z/ );

    my $url = '/includes/' . $listname;
    my ( $result, $code, $msg ) = $self->_make_rest_call( 'GET' => $url );

    # success
    return $result if ($result);

    # bad result
    return ( [ ERROR => $msg ] );
}

=head2 edit_includes 

 Usage    : my $result = $bt->edit_includes($listname, $listarray);
 Purpose  : Calls PUT /includes/:listname
 Argument : $listname = name of list with includes to edit
          : $listarray = array ref containing included listnames.
 Returns  : An arrayref containing the updated includes list.
          : [ ERROR, $msg ] on any errors.

=cut

sub edit_includes {
    my $self     = shift;
    my $listname = shift;
    my $aref     = shift || [];
    $aref = [ $aref, @_ ] if ( ref $aref ne 'ARRAY' );    # coerce arrayref

    return (
        { ERROR => "listname '$listname' contains unexpected characters" } )
        if ( $listname !~ /\A[\w\-]+\z/ );

    my $url = '/includes/' . $listname;
    my ( $result, $code, $msg )
        = $self->_make_rest_call( 'PUT' => $url, $aref );

    # success
    return $result if ($result);

    # bad result
    return ( [ ERROR => $msg ] );
}

# reads REST-like static json files for block/allow sets
# also handles recent lists
sub _get_db_static {
    my $self     = shift;
    my $type     = shift || 'bl';
    my $listname = shift || $self->{LIST};

    # whitelist deprecation patch
    $type = 'wl' if ($type eq 'al');

    my $cidrset = Net::CIDR::Set->new( { type => 'ipv4' } );

    if ( !$listname ) {
        carp "get_db: listname not set";
        return;
    }

    if ( $listname !~ /\A[\w\-]+\z/ ) {
        carp "get_db: listname '$listname' contains unexpected characters";
        return;
    }

    # sometimes it just can't connect. try more than once
    my $retry = 0;
    my $url   = '/static/' . $listname . '.' . $type . '.json';
    my ( $result, $code, $msg );
    while ( $retry < 3 ) {
        ( $result, $code, $msg ) = $self->_make_rest_call( 'GET' => $url );
        last if ( $code < 300 );
        $retry++;
        sleep(2);
    }
    if ( $code > 299 ) {
        carp "get_db $type: " . $msg;
        return;
    }

    # recent lists just return
    return $result if ( $listname eq 'recent' && $type =~ /\d/ );

    my $ct = 0;
    foreach my $ip (@$result) {
        $cidrset->add($ip);
        $ct++;
    }
    print STDERR "get_db $listname $type: read $ct entries\n"
        if ( $self->{DEBUG} );

    return $cidrset;
}

# common handler for REST calls
# returns ($result, $code, $msg) on success
sub _make_rest_call {
    my $self   = shift;
    my @params = @_;      # blindly pass args to RESTclient

    my $client = $self->{REST_CLIENT};
    return ( undef, 400, 'REST client did not initialize' ) if ( !$client );

    # make call, return on success
    $client->request(@params);
    my $code = $client->code;
    my $msg  = $client->status_line;
    $msg =~ s/\A\d+\s+//;    # strips code off
    my $result = $client->decode();
    print STDERR "$params[0] $params[1] => $code $msg\n"
        if ( $self->{DEBUG} );
    return ( $result, $code, $msg ) if ( $client->is_success );

    # Database error on server, abort the script
    if (   $code == 500
        && defined $result->{exception}
        && $result->{exception} =~ /DBI/ )
    {
        die "REST server DB is not available, abort!\n";
    }

    # Authorization error, abort the script
    if (   $code == 403
        && defined $result->{message}
        && $result->{message} =~ /auth check/ )
    {
        die "REST ID/KEY not accepted, check your config.\n";
    }

    # Missing Auth keyid, abort again
    if (   $code == 401
        && defined $result->{message}
        && $result->{message} =~ /unknown keyid/ )
    {
        die "REST ID/KEY not accepted, check your config.\n";
    }

    # other errors, pass back the code and message
    $msg = $result->{message} || $msg;
    return ( undef, $code, $msg );
}

# read module settings from config file
# any failures are quietly skipped
sub _read_config {
    my $self = shift;

    # try to find/open the config file
    my $config = $self->{CONFIG} || return;
    open( my $fh, '<', $config ) || return;

    # read in lines with keys that match existing self entries
    while ( my $lin = <$fh> ) {
        chomp $lin;
        next if ( $lin =~ /\A\s*(#.*)?\z/ );
        if ( $lin =~ m{(\w+)\s*=\s*(\S+)} ) {
            my ( $key, $val ) = ( uc($1), $2 );
            $self->{$key} = $val if ( defined $self->{$key} );
        }
    }
    close($fh);
}

# provide a default reported_by for rules
sub _default_reported_by {
    my $self = shift;

    my $userid   = $ENV{LOGNAME} || getpwuid($>) || 'user';
    my $hostname = hostname()    || 'a host';
    return "$userid at $hostname";
}

1;    # Magic true value required at end of module
__END__

=head1 DIAGNOSTICS

Most of these functions pass error messages directly from
the REST API that generates them. See the 
L<BadTraffic2 REST API|https://sysnews.ncsu.edu/restdocs/badtraffic2/>
documentation for details.

=head1 CONFIGURATION AND ENVIRONMENT

There is a sample configuration file found in the
etc/badtraffic.conf.example file in this distribution. You should copy
that file to /etc/badtraffic.conf and set the variables inside it
correctly for your server and REST API key.

You can set BT_CONFIG in your environment to point to an alternate
config file. Or you can specify an alternate path for that config file
using the (CONFIG => filepath) option. Or you can hard code the
information directly in your scripts if you prefer.

The installed scripts will look for the config file in that default
location, or you can specify and alternative using the -c option to
the script.

A security consideration - the config file will probably contain a copy
of your private API key. If this is installed in a world readable file
on your server, anyone on that server can use this library and the
helper scripts to modify the rules and lists with your key. It would
be best to restrict read access to the users that need to use the key.
Most likely mode 0600, owner root or an admin/cron account would be
sufficient.

=head1 DEPENDENCIES

=over

=item perl(JSON) from CPAN

=item perl(Net::CIDR::Set) from CPAN

=item perl(URI) from CPAN

=item perl(NCSU::RESTclient) version 0.5.1 or newer, from our Vision4 repository or from L<github.ncsu.edu|https://github.ncsu.edu/OIT-IWS/ncsu-restclient-perl5>.

=back

=head1 INCOMPATIBILITIES

NCSU::RESTclient version 0.4.x or lower will not encode POST/PUT 
requests to REST correctly. 

=head1 BUGS AND LIMITATIONS

No bugs have been reported.

Please report any bugs or feature requests though the github.ncsu.edu 
repository for this module.

=head1 AUTHOR

Charles J. Brabec  C<< <brabec@ncsu.edu> >>


=head1 LICENCE AND COPYRIGHT

Copyright (c) 2019, Charles J. Brabec C<< <brabec@ncsu.edu> >>. All rights reserved.

This module is free software; you can redistribute it and/or
modify it under the same terms as Perl itself. See L<perlartistic>.


=head1 DISCLAIMER OF WARRANTY

BECAUSE THIS SOFTWARE IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE SOFTWARE, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE SOFTWARE "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE SOFTWARE IS WITH
YOU. SHOULD THE SOFTWARE PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL
NECESSARY SERVICING, REPAIR, OR CORRECTION.

IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE SOFTWARE AS PERMITTED BY THE ABOVE LICENCE, BE
LIABLE TO YOU FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL,
OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE
THE SOFTWARE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING
RENDERED INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A
FAILURE OF THE SOFTWARE TO OPERATE WITH ANY OTHER SOFTWARE), EVEN IF
SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF
SUCH DAMAGES.
